@@ -50,6 +50,13 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
 # here and we'll add it to the list as required.
 _LOG_CONTENT_TYPES = set(['application/json'])
 
+_MAX_RETRY_INTERVAL = 60.0
+_EXPONENTIAL_DELAY_START = 0.5
+
+# NOTE(efried): This is defined in oslo_middleware.request_id.INBOUND_HEADER,
+# but it didn't seem worth adding oslo_middleware to requirements just for that
+_REQUEST_ID_HEADER = 'X-Openstack-Request-Id'
+
 
 def _construct_session(session_obj=None):
     # NOTE(morganfainberg): if the logic in this function changes be sure to
@@ -233,6 +240,29 @@ class RequestTiming(object):
         self.elapsed = elapsed
 
 
+class _Retries(object):
+    __slots__ = ('_fixed_delay', '_current')
+
+    def __init__(self, fixed_delay=None):
+        self._fixed_delay = fixed_delay
+        self.reset()
+
+    def __next__(self):
+        value = self._current
+        if not self._fixed_delay:
+            self._current = min(value * 2, _MAX_RETRY_INTERVAL)
+        return value
+
+    def reset(self):
+        if self._fixed_delay:
+            self._current = self._fixed_delay
+        else:
+            self._current = _EXPONENTIAL_DELAY_START
+
+    # Python 2 compatibility
+    next = __next__
+
+
 class Session(object):
     """Maintains client communication state and common functionality.
 
@@ -300,6 +330,9 @@ class Session(object):
     :param rate_semaphore: Semaphore to be used to control concurrency
                            and rate limiting of requests. (optional,
                            defaults to no concurrency or rate control)
+    :param int connect_retries: the maximum number of retries that should
+                                be attempted for connection errors.
+                                (optional, defaults to 0 - never retry).
     """
 
     user_agent = None
@@ -313,10 +346,18 @@ class Session(object):
                  redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
                  app_name=None, app_version=None, additional_user_agent=None,
                  discovery_cache=None, split_loggers=None,
-                 collect_timing=False, rate_semaphore=None):
+                 collect_timing=False, rate_semaphore=None,
+                 connect_retries=0):
 
         self.auth = auth
         self.session = _construct_session(session)
+        # NOTE(mwhahaha): keep a reference to the session object so we can
+        # clean it up when this object goes away. We don't want to close the
+        # session if it was passed into us as it may be reused externally.
+        # See LP#1838704
+        self._session = None
+        if not session:
+            self._session = self.session
         self.original_ip = original_ip
         self.verify = verify
         self.cert = cert
@@ -334,6 +375,7 @@ class Session(object):
         # so we can distinguish between the value being set or not.
         self._split_loggers = split_loggers
         self._collect_timing = collect_timing
+        self._connect_retries = connect_retries
         self._api_times = []
         self._rate_semaphore = rate_semaphore or NoOpSemaphore()
 
@@ -345,6 +387,17 @@ class Session(object):
 
         self._json = _JSONEncoder()
 
+    def __del__(self):
+        """Clean up resources on delete."""
+        if self._session:
+            # If we created a requests.Session, try to close it out correctly
+            try:
+                self._session.close()
+            except Exception:
+                pass
+            finally:
+                self._session = None
+
     @property
     def adapters(self):
         return self.session.adapters
@@ -573,11 +626,13 @@ class Session(object):
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
-                endpoint_override=None, connect_retries=0, logger=None,
+                endpoint_override=None, connect_retries=None, logger=None,
                 allow=None, client_name=None, client_version=None,
                 microversion=None, microversion_service_type=None,
                 status_code_retries=0, retriable_status_codes=None,
-                rate_semaphore=None, **kwargs):
+                rate_semaphore=None, global_request_id=None,
+                connect_retry_delay=None, status_code_retry_delay=None,
+                **kwargs):
         """Send an HTTP request with the specified characteristics.
 
         Wrapper around `requests.Session.request` to handle tasks such as
@@ -604,7 +659,7 @@ class Session(object):
                                   for forever/never. (optional)
         :param int connect_retries: the maximum number of retries that should
                                     be attempted for connection errors.
-                                    (optional, defaults to 0 - never retry).
+                                    (optional, defaults to None - never retry).
         :param bool authenticated: True if a token should be attached to this
                                    request, False if not or None for attach if
                                    an auth_plugin is available.
@@ -666,6 +721,17 @@ class Session(object):
         :param rate_semaphore: Semaphore to be used to control concurrency
                                and rate limiting of requests. (optional,
                                defaults to no concurrency or rate control)
+        :param global_request_id: Value for the X-Openstack-Request-Id header.
+        :param float connect_retry_delay: Delay (in seconds) between two
+                                          connect retries (if enabled).
+                                          By default exponential retry starting
+                                          with 0.5 seconds up to a maximum of
+                                          60 seconds is used.
+        :param float status_code_retry_delay: Delay (in seconds) between two
+                                              status code retries (if enabled).
+                                              By default exponential retry
+                                              starting with 0.5 seconds up to
+                                              a maximum of 60 seconds is used.
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -687,6 +753,8 @@ class Session(object):
         else:
             split_loggers = None
         logger = logger or utils.get_logger(__name__)
+        if connect_retries is None:
+            connect_retries = self._connect_retries
         # HTTP 503 - Service Unavailable
         retriable_status_codes = retriable_status_codes or [503]
         rate_semaphore = rate_semaphore or self._rate_semaphore
@@ -783,6 +851,12 @@ class Session(object):
             headers.setdefault('Content-Type', 'application/json')
             kwargs['data'] = self._json.encode(json)
 
+        if global_request_id is not None:
+            # NOTE(efried): This does *not* setdefault. If a global_request_id
+            # kwarg was explicitly specified, it should override any value
+            # previously configured (e.g. in Adapter.global_request_id).
+            headers[_REQUEST_ID_HEADER] = global_request_id
+
         for k, v in self.additional_headers.items():
             headers.setdefault(k, v)
 
@@ -814,11 +888,15 @@ class Session(object):
         if redirect is None:
             redirect = self.redirect
 
+        connect_retry_delays = _Retries(connect_retry_delay)
+        status_code_retry_delays = _Retries(status_code_retry_delay)
+
         send = functools.partial(self._send_request,
                                  url, method, redirect, log, logger,
                                  split_loggers, connect_retries,
                                  status_code_retries, retriable_status_codes,
-                                 rate_semaphore)
+                                 rate_semaphore, connect_retry_delays,
+                                 status_code_retry_delays)
 
         try:
             connection_params = self.get_auth_connection_params(auth=auth)
@@ -907,7 +985,7 @@ class Session(object):
     def _send_request(self, url, method, redirect, log, logger, split_loggers,
                       connect_retries, status_code_retries,
                       retriable_status_codes, rate_semaphore,
-                      connect_retry_delay=0.5, status_code_retry_delay=0.5,
+                      connect_retry_delays, status_code_retry_delays,
                       **kwargs):
         # NOTE(jamielennox): We handle redirection manually because the
         # requests lib follows some browser patterns where it will redirect
@@ -949,9 +1027,10 @@ class Session(object):
             if connect_retries <= 0:
                 raise
 
+            delay = next(connect_retry_delays)
             logger.info('Failure: %(e)s. Retrying in %(delay).1fs.',
-                        {'e': e, 'delay': connect_retry_delay})
-            time.sleep(connect_retry_delay)
+                        {'e': e, 'delay': delay})
+            time.sleep(delay)
 
             return self._send_request(
                 url, method, redirect, log, logger, split_loggers,
@@ -959,7 +1038,8 @@ class Session(object):
                 retriable_status_codes=retriable_status_codes,
                 rate_semaphore=rate_semaphore,
                 connect_retries=connect_retries - 1,
-                connect_retry_delay=connect_retry_delay * 2,
+                connect_retry_delays=connect_retry_delays,
+                status_code_retry_delays=status_code_retry_delays,
                 **kwargs)
 
         if log:
@@ -984,14 +1064,18 @@ class Session(object):
                 logger.warning("Failed to redirect request to %s as new "
                                "location was not provided.", resp.url)
             else:
-                # NOTE(jamielennox): We don't pass through connect_retry_delay.
+                # NOTE(jamielennox): We don't keep increasing delays.
                 # This request actually worked so we can reset the delay count.
+                connect_retry_delays.reset()
+                status_code_retry_delays.reset()
                 new_resp = self._send_request(
                     location, method, redirect, log, logger, split_loggers,
                     rate_semaphore=rate_semaphore,
                     connect_retries=connect_retries,
                     status_code_retries=status_code_retries,
                     retriable_status_codes=retriable_status_codes,
+                    connect_retry_delays=connect_retry_delays,
+                    status_code_retry_delays=status_code_retry_delays,
                     **kwargs)
 
                 if not isinstance(new_resp.history, list):
@@ -1001,21 +1085,23 @@ class Session(object):
         elif (resp.status_code in retriable_status_codes and
               status_code_retries > 0):
 
+            delay = next(status_code_retry_delays)
             logger.info('Retriable status code %(code)s. Retrying in '
                         '%(delay).1fs.',
-                        {'code': resp.status_code,
-                         'delay': status_code_retry_delay})
-            time.sleep(status_code_retry_delay)
+                        {'code': resp.status_code, 'delay': delay})
+            time.sleep(delay)
 
-            # NOTE(jamielennox): We don't pass through connect_retry_delay.
+            # NOTE(jamielennox): We don't keep increasing connection delays.
             # This request actually worked so we can reset the delay count.
+            connect_retry_delays.reset()
             return self._send_request(
                 url, method, redirect, log, logger, split_loggers,
                 connect_retries=connect_retries,
                 status_code_retries=status_code_retries - 1,
                 retriable_status_codes=retriable_status_codes,
                 rate_semaphore=rate_semaphore,
-                status_code_retry_delay=status_code_retry_delay * 2,
+                connect_retry_delays=connect_retry_delays,
+                status_code_retry_delays=status_code_retry_delays,
                 **kwargs)
 
         return resp
